from pathlib import Path
from Bio.Align import MultipleSeqAlignment
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import ElementTree
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq


class BeastXML:
    """Class for editing BEAST XML files."""

    file_name: Path
    traits: list
    xml: ElementTree  # what is?

    def __init__(
        self,
        file_name: Path,
        traits: list = [],
    ):
        self.file_name = file_name
        self.traits = traits
        self.xml = self._load_xml()
        # should check to see if there are traits in the xml
        # that are not in self.traits.

    def _load_xml(self):
        return ET.parse(self.file_name)

    @property
    def alignment(self) -> MultipleSeqAlignment:
        msa = MultipleSeqAlignment([])
        data = self.xml.find("data")
        for sequence_el in data:
            msa.append(
                SeqRecord(
                    Seq(sequence_el.get("value")),
                    id=sequence_el.get("taxon"),
                    description="",
                )
            )
        return msa

    def get_sequence_ids(self) -> list:
        return [s.id for s in self.alignment]

    def _add_trait_data(self, sequence_id, traitname, deliminator, group):
        trait_el = self.xml.find(f".//*[@traitname='{traitname}']")
        if not trait_el:
            raise ValueError(f"Could not find trait with traitname '{traitname}'")

        trait = sequence_id.split(deliminator)[group]
        trait_el.set("value", f"{trait_el.get('value')},{sequence_id}={trait}")

    def add_sequence(self, record: Seq):
        for trait in self.traits:
            self._add_trait_data(record.id, **trait)
        data = self.xml.find("data")
        sequence_el = ET.Element(
            "sequence",
            {
                "id": f"seq_{record.id}",
                "taxon": record.id,
                "totalcount": "4",
                "value": record.seq,
            },
        )
        data.append(sequence_el)

    def write(self, out_file=None) -> None:
        if not out_file:
            out_file = self.file_name
        self.xml.write(out_file)



from pathlib import Path
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.Align import MultipleSeqAlignment
import typer
from typing import List, Optional, Tuple

from .xml import BeastXML
from .state import StateFile

app = typer.Typer()


def find_closest_sequence(sequences_from_xml: MultipleSeqAlignment, new_sequence: Seq):
    max_score = None
    seq_id = None
    with typer.progressbar(sequences_from_xml) as progress:
        for i, sequence in enumerate(progress):
            score = sum(
                xi != yi for xi, yi in zip(str(sequence.seq), str(new_sequence))
            )
            if not max_score:
                max_score = score
                seq_id = i
            elif score < max_score:
                max_score = score
                seq_id = i
    if seq_id == "None":
        raise Exception("No Seq found?")
    return seq_id, max_score


def get_sequences_to_add(fasta_file, list_of_seq_ids: list):
    records = SeqIO.parse(fasta_file, "fasta")
    return [record for record in records if record.id not in list_of_seq_ids]


@app.command()
def main(
    xml_file: Path,
    fasta_file: Path,
    state_file: Path = None,
    output: Path = None,
    trait: Optional[List[str]] = typer.Option(
        None,
        help="Trait information 'traitname deliminator group' string seperated by spaces",
    ),
):
    if not state_file:
        state_file = Path(f"{xml_file}.state")
    state_tree = StateFile(state_file)
    traits = [
        {
            "traitname": t.split(" ")[0],
            "deliminator": t.split(" ")[0],
            "group": int(t.split(" ")[2]),
        }
        for t in trait
    ]

    beast_xml = BeastXML(xml_file, traits)

    sequences_to_add = get_sequences_to_add(fasta_file, beast_xml.get_sequence_ids())

    if not sequences_to_add:
        typer.echo("No new sequences found in the fasta file.")
        raise typer.Exit(code=1)

    for sequence in sequences_to_add:
        typer.echo(f"Adding new sequence: {sequence.id}")
        if len(sequence) != beast_xml.alignment.get_alignment_length():
            raise ValueError("Sequences must all be the same length")
        closest_seq_id, max_score = find_closest_sequence(
            beast_xml.alignment, sequence.seq
        )
        new_clade = state_tree.graft(closest_seq_id)
        name = new_clade.name
        new_clade.name = sequence.id
        state_tree.draw()
        new_clade.name = name
        beast_xml.add_sequence(sequence)

    beast_xml.write(out_file=output)
    state_tree.write(out_file=output)



from pathlib import Path
from Bio import Phylo
from io import StringIO


class StateFile:
    """Class for editing BEAST state files."""

    file_name: str
    tree: str

    def __init__(self, file_name: Path):
        self.file_name = file_name
        self.tree = self._get_tree()

    def _get_tree(self):
        with open(self.file_name) as f:
            lines = f.readlines()
        tree_line = lines[1]
        newick_tree = tree_line.split(">")[1].split("</")[0]
        return Phylo.read(StringIO(newick_tree), "newick")

    @property
    def newick(self):
        writer = Phylo.NewickIO.Writer([self.tree])
        return next(writer.to_strings(format_branch_length="%1.17f"))

    def graft(self, branch_id, graft_point: float = 0.5, branch_length: float = None):
        clade = next(c for c in self.tree.get_terminals() if c.name == str(branch_id))
        if not clade:
            raise ValueError("Could not find branch with id", branch_id)
        total_branch_length = clade.branch_length
        clade.branch_length = total_branch_length * graft_point
        clade.split(branch_length=total_branch_length * (1 - graft_point))
        clade.confidence = 1
        clade.name = None
        original_branch, new_branch = clade.clades
        original_branch.name = str(branch_id)
        new_branch.name = str(len(self.tree.get_terminals()) - 1)
        if branch_length:
            new_branch.branch_length = branch_length
        return new_branch

    def draw(self):
        Phylo.draw_ascii(self.tree)

    def write(self, out_file=None) -> None:
        with open(self.file_name) as f:
            state_file_lines = f.readlines()

        old_tree_line = state_file_lines[1]
        opening_tag = old_tree_line.split(">")[0]
        closting_tag = old_tree_line.split("</")[-1]
        state_file_lines[1] = f"{opening_tag}>{self.newick}</{closting_tag}"

        if not out_file:
            out_file = self.file_name
        else:
            out_file = Path(f"{out_file}.state")

        with open(out_file, "w") as f:
            f.writelines(state_file_lines)
